@@ -17,7 +17,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
  *
- * $Id: mms.c,v 1.21 2006/07/13 12:32:20 shawarma Exp $
+ * $Id: mms.c,v 1.31 2007/12/11 20:35:01 jwrdegoede Exp $
  *
  * MMS over TCP protocol
  *   based on work from major mms
@@ -61,6 +61,7 @@
 # define lprintf(x...)
 #endif
 
+#define __MMS_C__
 
 #include "bswap.h"
 #include "mms.h"
@@ -78,9 +79,9 @@
 
 #define CMD_HEADER_LEN   40
 #define CMD_PREFIX_LEN    8
-#define CMD_BODY_LEN   1024
+#define CMD_BODY_LEN   1024 * 16 /* FIXME: make this dynamic */
 
-#define ASF_HEADER_LEN 8192
+#define ASF_HEADER_LEN (8192 * 2)
 
 
 #define MMS_PACKET_ERR        0
@@ -106,6 +107,13 @@ struct mms_packet_header_s {
   uint32_t  packet_seq;
 };
 
+typedef struct mms_stream_s mms_stream_t;
+struct mms_stream_s {
+  int           stream_id;
+  int           stream_type;
+  uint32_t      bitrate;
+  uint32_t      bitrate_pos;
+};
 
 struct mms_s {
 
@@ -134,26 +142,30 @@ struct mms_s {
   uint8_t       buf[BUF_SIZE];
   int           buf_size;
   int           buf_read;
+  off_t         buf_packet_seq_offset; /* packet sequence offset residing in
+                                          buf */
   
   uint8_t       asf_header[ASF_HEADER_LEN];
   uint32_t      asf_header_len;
   uint32_t      asf_header_read;
   int           seq_num;
   int           num_stream_ids;
-  int           stream_ids[ASF_MAX_NUM_STREAMS];
-  int           stream_types[ASF_MAX_NUM_STREAMS];
+  mms_stream_t  streams[ASF_MAX_NUM_STREAMS];
+  uint8_t       packet_id_type;
   off_t         start_packet_seq; /* for live streams != 0, need to keep it around */
   int           need_discont; /* whether we need to set start_packet_seq */
   uint32_t      asf_packet_len;
   uint64_t      file_len;
+  uint64_t      time_len; /* playback time in 100 nanosecs (10^-7) */
+  uint64_t      preroll;
+  uint64_t      asf_num_packets;
   char          guid[37];
-  uint32_t      bitrates[ASF_MAX_NUM_STREAMS];
-  uint32_t      bitrates_pos[ASF_MAX_NUM_STREAMS];
   int           bandwidth;
   
   int           has_audio;
   int           has_video;
   int           live_flag;
+  int           seekable;
   off_t         current_pos;
   int           eos;
 };
@@ -528,6 +540,7 @@ static int get_packet_header (mms_io_t *io, mms_t *this, mms_packet_header_t *he
   header->flags          = 0;
   header->packet_id_type = 0;
   len = io_read(io,  this->s, this->buf, 8);
+  this->buf_packet_seq_offset = -1;
   if (len != 8)
     goto error;
 
@@ -578,6 +591,7 @@ static int get_packet_command (mms_io_t *io, mms_t *this, uint32_t packet_len) {
   lprintf("packet_len: %d bytes\n", packet_len);
 
   len = io_read(io,  this->s, this->buf + 12, packet_len) ;
+  //this->buf_packet_seq_offset = -1; // already set in get_packet_header
   if (len != packet_len) {
     return 0;
   }
@@ -680,7 +694,8 @@ static int get_asf_header (mms_io_t *io, mms_t *this) {
       case MMS_PACKET_ASF_PACKET:
         if (header.packet_len + this->asf_header_len > ASF_HEADER_LEN) {
             lprintf( "***LOG:*** -- "
-                     "libmms: asf packet too large\n");
+                     "libmms: asf packet too large: %d\n", 
+                     header.packet_len + this->asf_header_len);
             return 0;
         }
         len = io_read(io,  this->s,
@@ -708,6 +723,7 @@ static void interp_asf_header (mms_t *this) {
 
   this->asf_packet_len = 0;
   this->num_stream_ids = 0;
+  this->asf_num_packets = 0;
   /*
    * parse header
    */
@@ -736,6 +752,9 @@ static void interp_asf_header (mms_t *this) {
           break;
         }
         this->file_len       = LE_64(this->asf_header + i + 40 - 24);
+        this->time_len       = LE_64(this->asf_header + i + 64 - 24);
+        //this->time_len       = LE_64(this->asf_header + i + 72 - 24);
+        this->preroll        = LE_64(this->asf_header + i + 80 - 24);
         lprintf ("file object, packet length = %d (%d)\n",
                  this->asf_packet_len, LE_32(this->asf_header + i + 96 - 24));
         break;
@@ -776,9 +795,9 @@ static void interp_asf_header (mms_t *this) {
           lprintf ("stream object, stream id: %d, type: %d, encrypted: %d\n",
 		   stream_id, type, encrypted);
 
-	  if (this->num_stream_ids < ASF_MAX_NUM_STREAMS && stream_id < ASF_MAX_NUM_STREAMS) {
-	    this->stream_types[stream_id] = type;
-	    this->stream_ids[this->num_stream_ids] = stream_id;
+	  if (this->num_stream_ids < ASF_MAX_NUM_STREAMS) {
+	    this->streams[this->num_stream_ids].stream_type = type;
+	    this->streams[this->num_stream_ids].stream_id = stream_id;
 	    this->num_stream_ids++;
 	  } else {
 	    lprintf ("too many streams, skipping\n");
@@ -796,16 +815,27 @@ static void interp_asf_header (mms_t *this) {
           lprintf ("streams %d\n", streams); 
 
           for(j = 0; j < streams; j++) {
+            int stream_index;
             stream_id = LE_16(this->asf_header + i + 2 + j * 6);
-            lprintf ("stream id %d\n", stream_id); 
-            this->bitrates[stream_id] = LE_32(this->asf_header + i + 4 + j * 6);
-            this->bitrates_pos[stream_id] = i + 4 + j * 6;
-            lprintf ("stream id %d, bitrate %d\n", stream_id, 
-                     this->bitrates[stream_id]);
+            lprintf ("stream id %d\n", stream_id);
+            for(stream_index = 0; stream_index < this->num_stream_ids; stream_index++) {
+              if (this->streams[stream_index].stream_id == stream_id)
+                break;
+            }
+            if (stream_index < this->num_stream_ids) {
+              this->streams[stream_index].bitrate = LE_32(this->asf_header + i + 4 + j * 6);
+              this->streams[stream_index].bitrate_pos = i + 4 + j * 6;
+              lprintf ("stream id %d, bitrate %d\n", stream_id, 
+                     this->[stream_index].bitrate);
+            }
           }
         }
         break;
     
+      case GUID_ASF_DATA:
+        this->asf_num_packets = LE_64(this->asf_header + i + 40 - 24);
+        break;
+
       default:
         lprintf ("unknown object\n");
         break;
@@ -913,7 +943,6 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
   int     max_arate    = 0;
   int     min_vrate    = 0;
   int     min_bw_left  = 0;
-  int     stream_id;
   int     bandwitdh_left;
   int     res;
 
@@ -922,12 +951,11 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
   /* i've never seen more than one audio stream */
   lprintf("num_stream_ids=%d\n", this->num_stream_ids);
   for (i = 0; i < this->num_stream_ids; i++) {
-    stream_id = this->stream_ids[i];
-    switch (this->stream_types[stream_id]) {
+    switch (this->streams[i].stream_type) {
       case ASF_STREAM_TYPE_AUDIO:
-        if (this->bitrates[stream_id] > max_arate) {
-          audio_stream = stream_id;
-          max_arate = this->bitrates[stream_id];
+        if (this->streams[i].bitrate > max_arate) {
+          audio_stream = this->streams[i].stream_id;
+          max_arate = this->streams[i].bitrate;
         }
         break;
       default:
@@ -944,13 +972,12 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
 
   min_bw_left = bandwitdh_left;
   for (i = 0; i < this->num_stream_ids; i++) {
-    stream_id = this->stream_ids[i];
-    switch (this->stream_types[stream_id]) {
+    switch (this->streams[i].stream_type) {
       case ASF_STREAM_TYPE_VIDEO:
-        if (((bandwitdh_left - this->bitrates[stream_id]) < min_bw_left) &&
-            (bandwitdh_left >= this->bitrates[stream_id])) {
-          video_stream = stream_id;
-          min_bw_left = bandwitdh_left - this->bitrates[stream_id];
+        if (((bandwitdh_left - this->streams[i].bitrate) < min_bw_left) &&
+            (bandwitdh_left >= this->streams[i].bitrate)) {
+          video_stream = this->streams[i].stream_id;
+          min_bw_left = bandwitdh_left - this->streams[i].bitrate;
         }
         break;
       default:
@@ -961,13 +988,12 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
   /* choose the lower bitrate of */
   if (!video_stream && this->has_video) {
     for (i = 0; i < this->num_stream_ids; i++) {
-      stream_id = this->stream_ids[i];
-      switch (this->stream_types[stream_id]) {
+      switch (this->streams[i].stream_type) {
         case ASF_STREAM_TYPE_VIDEO:
-          if ((this->bitrates[stream_id] < min_vrate) ||
+          if ((this->streams[i].bitrate < min_vrate) ||
               (!min_vrate)) {
-            video_stream = stream_id;
-            min_vrate = this->bitrates[stream_id];
+            video_stream = this->streams[i].stream_id;
+            min_vrate = this->streams[i].bitrate;
           }
           break;
         default:
@@ -982,10 +1008,10 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
   for (i = 1; i < this->num_stream_ids; i++) {
     this->scmd_body [ (i - 1) * 6 + 2 ] = 0xFF;
     this->scmd_body [ (i - 1) * 6 + 3 ] = 0xFF;
-    this->scmd_body [ (i - 1) * 6 + 4 ] = this->stream_ids[i] ;
-    this->scmd_body [ (i - 1) * 6 + 5 ] = this->stream_ids[i] >> 8;
-    if ((this->stream_ids[i] == audio_stream) ||
-        (this->stream_ids[i] == video_stream)) {
+    this->scmd_body [ (i - 1) * 6 + 4 ] = this->streams[i].stream_id ;
+    this->scmd_body [ (i - 1) * 6 + 5 ] = this->streams[i].stream_id >> 8;
+    if ((this->streams[i].stream_id == audio_stream) ||
+        (this->streams[i].stream_id == video_stream)) {
       this->scmd_body [ (i - 1) * 6 + 6 ] = 0x00;
       this->scmd_body [ (i - 1) * 6 + 7 ] = 0x00;
     } else {
@@ -994,17 +1020,17 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
       this->scmd_body [ (i - 1) * 6 + 7 ] = 0x00;
       
       /* forces the asf demuxer to not choose this stream */
-      if (this->bitrates_pos[this->stream_ids[i]]) {
-        this->asf_header[this->bitrates_pos[this->stream_ids[i]]]     = 0;
-        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 1] = 0;
-        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 2] = 0;
-        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 3] = 0;
+      if (this->streams[i].bitrate_pos) {
+        this->asf_header[this->streams[i].bitrate_pos    ] = 0;
+        this->asf_header[this->streams[i].bitrate_pos + 1] = 0;
+        this->asf_header[this->streams[i].bitrate_pos + 2] = 0;
+        this->asf_header[this->streams[i].bitrate_pos + 3] = 0;
       }
     }
   }
 
   if (!send_command (io, this, 0x33, this->num_stream_ids, 
-                     0xFFFF | this->stream_ids[0] << 16, 
+                     0xFFFF | this->streams[0].stream_id << 16, 
                      this->num_stream_ids * 6 + 2)) {
   /* FIXME: de-xine-ification */
     lprintf ( "***LOG:*** -- "
@@ -1055,6 +1081,7 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
   this->need_discont    = 1;
   this->buf_size        = 0;
   this->buf_read        = 0;
+  this->buf_packet_seq_offset = -1;
   this->has_audio       = 0;
   this->has_video       = 0;
   this->bandwidth       = bandwidth;
@@ -1074,8 +1101,11 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
   this->host = uri->hostname;
   this->port = uri->port;
   this->password = uri->passwd;
-  this->uri = uri->path;
-  
+  this->uri = gnet_mms_helper(uri);
+
+  if(!this->uri)
+	goto fail;
+
   if (!mmst_valid_proto(this->proto)) {
     lprintf ("unsupported protocol\n");
     goto fail;
@@ -1158,6 +1188,13 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
     mms_buffer_init(&command_buffer, this->scmd_body);
     mms_buffer_put_32 (&command_buffer, 0x00000000); /* ?? */
     mms_buffer_put_32 (&command_buffer, 0x00000000); /* ?? */
+
+    /* FIXME: refuse to work with urls that are longer that buffer can hold
+       64 is a precation */
+
+    if (strlen(this->uri) >= CMD_BODY_LEN - 64)
+	goto fail;
+
     string_utf16 (url_conv, this->scmd_body + command_buffer.pos, this->uri, strlen(this->uri));
     if (!send_command (io, this, 5, 1, 0xffffffff, strlen(this->uri) * 2 + 12))
       goto fail;
@@ -1173,6 +1210,7 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
         xx = this->buf[62];
         yy = this->buf[63];
         this->live_flag = ((xx == 0) && ((yy & 0xf) == 2));
+        this->seekable = !this->live_flag;
         lprintf("live: live_flag=%d, xx=%d, yy=%d\n", this->live_flag, xx, yy);
       }
       break;
@@ -1229,6 +1267,8 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
     goto fail;
 
   interp_asf_header (this);
+  if (!this->asf_packet_len || !this->num_stream_ids)
+    goto fail;
 
   /* FIXME de-xine-ification */
 /*   report_progress (stream, 70); */
@@ -1244,6 +1284,7 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
 /*   report_progress (stream, 80); */
 
   /* command 0x07 */
+  this->packet_id_type = ASF_MEDIA_PACKET_ID_TYPE;
   {
     mms_buffer_t command_buffer;
     mms_buffer_init(&command_buffer, this->scmd_body);
@@ -1255,7 +1296,7 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
     mms_buffer_put_8  (&command_buffer, 0xFF);
     mms_buffer_put_8  (&command_buffer, 0xFF);
     mms_buffer_put_8  (&command_buffer, 0x00);                        /* stream time limit flag */
-    mms_buffer_put_32 (&command_buffer, ASF_MEDIA_PACKET_ID_TYPE);    /* asf media packet id type */
+    mms_buffer_put_32 (&command_buffer, this->packet_id_type);    /* asf media packet id type */
     if (!send_command (io, this, 0x07, 1, 0x0001FFFF, command_buffer.pos)) {
   /* FIXME: de-xine-ification */
       lprintf ( "***LOG:*** -- "
@@ -1340,6 +1381,8 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
               }
 
               interp_asf_header (this);
+              if (!this->asf_packet_len || !this->num_stream_ids)
+                return 0;
 
               if (!mms_choose_best_streams(io, this))
                 return 0;
@@ -1367,6 +1410,11 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
                 }
               }
               this->current_pos = 0;
+              
+              /* I don't know if this ever happens with none live (and thus
+                 seekable streams), but I do know that if it happens all bets
+                 with regards to seeking are off */
+              this->seekable = 0;
             }
             break;
 
@@ -1406,7 +1454,8 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
 
 	/* FIXME: probably needs some more sophisticated logic, but
 	   until we do seeking, this should work */
-	if(this->need_discont)
+	if(this->need_discont &&
+           header.packet_id_type == ASF_MEDIA_PACKET_ID_TYPE)
 	{
 	  this->need_discont = 0;
 	  this->start_packet_seq = header.packet_seq;
@@ -1443,11 +1492,22 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
 	      (start < end)  && (end < (base+BUF_SIZE-1))) {
 	    memset(this->buf + header.packet_len, 0, this->asf_packet_len - header.packet_len);
  	  }
+          if (header.packet_id_type == this->packet_id_type) {
 	  if (this->asf_packet_len > BUF_SIZE) {
             this->buf_size = BUF_SIZE;
           } else {
             this->buf_size = this->asf_packet_len;
           }
+            this->buf_packet_seq_offset =
+              header.packet_seq - this->start_packet_seq;
+          } else {
+            this->buf_size = 0;
+            // Don't set this packet sequence for reuse in seek(), since the
+            // subsequence packet may be discontinued.
+            //this->buf_packet_seq_offset = header.packet_seq;
+            // already set to -1 in get_packet_header
+            //this->buf_packet_seq_offset = -1;
+          }
 	}
       }
       break;
@@ -1521,6 +1581,261 @@ int mms_read (mms_io_t *io, mms_t *this, char *data, int len) {
   return total;
 }
 
+// To be inline function?
+static int mms_request_data_packet (mms_io_t *io, mms_t *this,
+  double time_sec, unsigned long first_packet, unsigned long time_msec_limit) {
+  /* command 0x07 */
+  {
+    mms_buffer_t command_buffer;
+    //mms_buffer_init(&command_buffer, this->scmd_body);
+    //mms_buffer_put_32 (&command_buffer, 0x00000000);                  /* 64 byte float timestamp */
+    //mms_buffer_put_32 (&command_buffer, 0x00000000);                  
+    memcpy(this->scmd_body, &time_sec, 8);
+    mms_buffer_init(&command_buffer, this->scmd_body+8);
+    mms_buffer_put_32 (&command_buffer, 0xFFFFFFFF);                  /* ?? */
+    //mms_buffer_put_32 (&command_buffer, 0xFFFFFFFF);                  /* first packet sequence */
+    mms_buffer_put_32 (&command_buffer, first_packet);                  /* first packet sequence */
+    //mms_buffer_put_8  (&command_buffer, 0xFF);                        /* max stream time limit (3 bytes) */
+    //mms_buffer_put_8  (&command_buffer, 0xFF);
+    //mms_buffer_put_8  (&command_buffer, 0xFF);
+    //mms_buffer_put_8  (&command_buffer, 0x00);                        /* stream time limit flag */
+    mms_buffer_put_32 (&command_buffer, time_msec_limit & 0x00FFFFFF);/* max stream time limit (3 bytes) */
+    mms_buffer_put_32 (&command_buffer, this->packet_id_type);    /* asf media packet id type */
+    if (!send_command (io, this, 0x07, 1, 0x0001FFFF, 8+command_buffer.pos)) {
+  /* FIXME: de-xine-ification */
+      lprintf ( "***LOG:*** -- "
+	       "libmms: failed to send command 0x07\n");
+      return 0;
+    }
+  }
+  /* TODO: adjust current_pos, considering asf_header_read */
+  return 1;
+}
+
+int mms_request_time_seek (mms_io_t *io, mms_t *this, double time_sec) {
+  if (++this->packet_id_type <= ASF_MEDIA_PACKET_ID_TYPE)
+    this->packet_id_type = ASF_MEDIA_PACKET_ID_TYPE+1;
+  //return mms_request_data_packet (io, this, time_sec, 0xFFFFFFFF, 0x00FFFFFF);
+  // also adjust time by preroll
+  return mms_request_data_packet (io, this,
+                                  time_sec+(double)(this->preroll)/1000,
+                                  0xFFFFFFFF, 0x00FFFFFF);
+}
+
+// set current_pos to the first byte of the requested packet by peeking at
+// the first packet.
+// To be inline function?
+static int peek_and_set_pos (mms_io_t *io, mms_t *this) {
+  uint8_t saved_buf[BUF_SIZE];
+  int     saved_buf_size;
+  off_t   saved_buf_packet_seq_offset;
+  // save buf and buf_size that may be changed in get_media_packet()
+  memcpy(saved_buf, this->buf, this->buf_size);
+  saved_buf_size = this->buf_size;
+  saved_buf_packet_seq_offset = this->buf_packet_seq_offset;
+  //this->buf_size = this->buf_read = 0; // reset buf, only if success peeking
+  this->buf_size = 0;
+  while (!this->eos) {
+    // get_media_packet() will set current_pos if data packet is read.
+    if (!get_media_packet (io, this)) {
+  /* FIXME: de-xine-ification */
+      lprintf ( "***LOG:*** -- "
+               "libmms: get_media_packet failed\n");
+      // restore buf and buf_size that may be changed in get_media_packet()
+      memcpy(this->buf, saved_buf, saved_buf_size);
+      this->buf_size = saved_buf_size;
+      this->buf_packet_seq_offset = saved_buf_packet_seq_offset;
+      return 0;
+    }
+    if (this->buf_size > 0) break;
+  }
+  // flush header and reset buf_read, only if success peeking
+  this->asf_header_read = this->asf_header_len;
+  this->buf_read = 0;
+  return 1;
+  //return this->current_pos;
+}
+
+// send time seek request, and update current_pos corresponding to the next
+// requested packet
+// Note that, the current_pos will always does not less than asf_header_len
+int mms_time_seek (mms_io_t *io, mms_t *this, double time_sec) {
+  if (!this->seekable)
+    return 0;
+
+  if (!mms_request_time_seek (io, this, time_sec)) return 0;
+  return peek_and_set_pos (io, this);
+}
+
+// http://sdp.ppona.com/zipfiles/MMSprotocol_pdf.zip said that, this
+// packet_seq value make no difference in version 9 servers.
+// But from my experiment with
+// mms://202.142.200.130/tltk/56k/tltkD2006-08-08ID-7209.wmv and
+// mms://202.142.200.130/tltk/56k/tltkD2006-09-01ID-7467.wmv (the url may valid
+// in only 2-3 months) whose server is version 9, it does response and return
+// the requested packet.
+int mms_request_packet_seek (mms_io_t *io, mms_t *this,
+                             unsigned long packet_seq) {
+  if (++this->packet_id_type <= ASF_MEDIA_PACKET_ID_TYPE)
+    this->packet_id_type = ASF_MEDIA_PACKET_ID_TYPE+1;
+  return mms_request_data_packet (io, this, 0, packet_seq, 0x00FFFFFF);
+}
+
+// send packet seek request, and update current_pos corresponding to the next
+// requested packet
+// Note that, the current_pos will always does not less than asf_header_len
+// Not export this function.  Let user use mms_seek() instead?
+static int mms_packet_seek (mms_io_t *io, mms_t *this,
+                            unsigned long packet_seq) {
+  if (!mms_request_packet_seek (io, this, packet_seq)) return 0;
+  return peek_and_set_pos (io, this);
+}
+
+/*
+TODO: To use this table to calculate buf_packet_seq_offset rather than store
+and retrieve it from this->buf_packet_seq_offset?
+current_packet_seq == (current_pos - asf_header_len) / asf_packet_len
+current_packet_seq == -1 if current_pos < asf_header_len
+buf_packet_seq_offset indicating which packet sequence are residing in the buf.
+Possible status after read(), "last" means last value or unchange.
+current_packet_seq | buf_read       | buf_size  | buf_packet_seq_offset
+-------------------+----------------+-----------+---------------
+<= 0               | 0 (last)       | 0 (last)  | none
+<= 0               | 0 (last)       | 0 (last)  | eos at #0
+<= 0               | 0 (last)       | 0 (last)  | eos at > #0
+<= 0               | 0 (last)       | > 0 (last)| #0
+<= 0               | buf_size (last)| > 0 (last)| > #0
+> 0                | 0              | 0         | eos at current_packet_seq
+> 0                | 0(never happen)| > 0       | (never happen)
+> 0                | buf_size       | > 0       | current_packet_seq-1
+*/
+// TODO: How to handle seek() in multi stream source?
+// The stream that follows 0x20 ("new stream") command.
+off_t mms_seek (mms_io_t *io, mms_t *this, off_t offset, int origin) {
+  off_t dest;
+  off_t dest_packet_seq;
+  //off_t buf_packet_seq_offset;
+  
+  if (!this->seekable)
+    return this->current_pos;
+  
+  switch (origin) {
+    case SEEK_SET:
+      dest = offset;
+      break;
+    case SEEK_CUR:
+      dest = this->current_pos + offset;
+      break;
+    case SEEK_END:
+      //if (this->asf_num_packets == 0) {
+      //  //printf ("input_mms: unknown end position in seek!\n");
+      //  return this->current_pos;
+      //}
+      dest = mms_get_length (this) + offset;
+    default:
+      printf ("input_mms: unknown origin in seek!\n");
+      return this->current_pos;
+  }
+
+  dest_packet_seq = dest - this->asf_header_len;
+  //if (dest_packet_seq > 0) dest_packet_seq /= this->asf_packet_len;
+  dest_packet_seq = dest_packet_seq >= 0 ?
+    dest_packet_seq / this->asf_packet_len : -1;
+#if 0
+  // buf_packet_seq_offset will identify which packet sequence are residing in
+  // the buf.
+#if 1 /* To show both of the alternate styles :D */
+  buf_packet_seq_offset = this->current_pos - this->asf_header_len;
+  //if (buf_packet_seq_offset > 0) buf_packet_seq_offset /= this->asf_packet_len;
+  buf_packet_seq_offset = buf_packet_seq_offset >= 0 ?
+    buf_packet_seq_offset / this->asf_packet_len : -1;
+  // Note: buf_read == buf_size == 0 may means that it is eos,
+  //       eos means that the packet has been peek'ed.
+  if (this->buf_read >= this->buf_size && this->buf_size > 0 &&
+      buf_packet_seq_offset >= 0 ||
+      // assuming packet not peek'ed in the following condition
+      /*this->buf_read >= this->buf_size && */this->buf_size == 0 &&
+      buf_packet_seq_offset == 0)
+    // The buf is all read but the packet has not been peek'ed.
+    --buf_packet_seq_offset;
+#else
+  buf_packet_seq_offset = this->current_pos - this->asf_header_len - 1;
+  //if (buf_packet_seq_offset > 0) buf_packet_seq_offset /= this->asf_packet_len;
+  buf_packet_seq_offset = buf_packet_seq_offset >= 0 ?
+    buf_packet_seq_offset / this->asf_packet_len : -1;
+  // Note: buf_read == buf_size == 0 may means that it is eos,
+  //       eos means that the packet has been peek'ed.
+  if (this->buf_read == 0/* && buf_packet_seq_offset >= 0*/)
+    // Since the packet has just been peek'ed.
+    ++buf_packet_seq_offset;
+#endif
+#endif
+
+  if (dest_packet_seq < 0) {
+    if (this->buf_packet_seq_offset > 0) {
+      if (!mms_request_packet_seek (io, this, 0xFFFFFFFF))
+        return this->current_pos;
+#if 1
+      // clear buf
+      this->buf_read = this->buf_size = 0;
+      this->buf_packet_seq_offset = -1;
+    } else {
+#else
+      // clear buf
+      this->buf_read = this->buf_size;
+      // Set this packet sequence not to be reused, since the subsequence
+      // packet may be discontinued.
+      this->buf_packet_seq_offset = -1;
+    // don't reset buf_read if buf_packet_seq_offset < 0, since the previous
+    // buf may not be cleared.
+    } else if (this->buf_packet_seq_offset == 0) {
+#endif
+      // reset buf_read
+      this->buf_read = 0;
+    }
+    this->asf_header_read = dest;
+    return this->current_pos = dest;
+  }
+  // dest_packet_seq >= 0
+  if (this->asf_num_packets && dest == this->asf_header_len +
+      this->asf_num_packets*this->asf_packet_len) {
+    // Requesting the packet beyond the last packet, can cause the server to
+    // not return any packet or any eos command.  This can cause
+    // mms_packet_seek() to hang.
+    // This is to allow seeking at end of stream, and avoid hanging.
+    --dest_packet_seq;
+  }
+  if (dest_packet_seq != this->buf_packet_seq_offset) {
+    if (this->asf_num_packets && dest_packet_seq >= this->asf_num_packets) {
+      // Do not seek beyond the last packet.
+      return this->current_pos;
+    }
+    if (!mms_packet_seek (io, this, this->start_packet_seq + dest_packet_seq))
+      return this->current_pos;
+    // Check if current_pos is correct.
+    // This can happen if the server ignore packet seek command.
+    // If so, just return unchanged current_pos, rather than trying to
+    // mms_read() to reach the destination pos.
+    // It should let the caller to decide to choose the alternate method, such
+    // as, mms_time_seek() and/or mms_read() until the destination pos is
+    // reached.
+    if (dest_packet_seq != this->buf_packet_seq_offset)
+      return this->current_pos;
+    // This has already been set in mms_packet_seek().
+    //if (current_packet_seq < 0)
+    //  this->asf_header_read = this->asf_header_len;
+    //this->asf_header_read = this->asf_header_len;
+  }
+  // eos is reached ?
+  //if (this->buf_size <= 0) return this->current_pos;
+  //this->buf_read = (dest - this->asf_header_len) % this->asf_packet_len;
+  this->buf_read = dest -
+    (this->asf_header_len + dest_packet_seq*this->asf_packet_len);
+  // will never happen ?
+  //if (this->buf_size <= this->buf_read) return this->current_pos;
+  return this->current_pos = dest;
+}
+
 
 void mms_close (mms_t *this) {
 
@@ -1542,10 +1857,38 @@ void mms_close (mms_t *this) {
   free (this);
 }
 
+double mms_get_time_length (mms_t *this) {
+  return (double)(this->time_len) / 1e7;
+}
+
+uint64_t mms_get_raw_time_length (mms_t *this) {
+  return this->time_len;
+}
+
 uint32_t mms_get_length (mms_t *this) {
-  return this->file_len;
+  /* we could / should return this->file_len here, but usually this->file_len
+     is longer then the calculation below, as usually an asf file contains an
+     asf index object after the data stream. However since we do not have a
+     (known) way to get to this index object through mms, we return a
+     calculated size of what we can get to when we know. */
+  if (this->asf_num_packets)
+    return this->asf_header_len + this->asf_num_packets*this->asf_packet_len;
+  else
+    return this->file_len;
 }
 
 off_t mms_get_current_pos (mms_t *this) {
   return this->current_pos;
 }
+
+uint32_t mms_get_asf_header_len (mms_t *this) {
+  return this->asf_header_len;
+}
+
+uint64_t mms_get_asf_packet_len (mms_t *this) {
+  return this->asf_packet_len;
+}
+
+int mms_get_seekable (mms_t *this) {
+  return this->seekable;
+}
